@@ -28,12 +28,15 @@ import Data.List
 import Data.Maybe
 import Data.Char (toUpper, toLower, isAlphaNum)
 import qualified Data.Map as Map
-import System.IO
+import Prelude hiding (writeFile, putStrLn)
+import System.IO.UTF8 (writeFile, putStrLn)
 import System.Directory
 import System.Environment
 import System.Exit
 import System.FilePath
 import Text.PrettyPrint
+import Text.Printf (printf)
+import Data.Char (ord)
 import Text.Highlighting.Kate.Definitions
 
 data SyntaxDefinition =
@@ -80,7 +83,6 @@ data SyntaxParser =
                , parserChildren          :: [SyntaxParser]
                } deriving (Read, Show)
 
-
 -- | Converts a list of files (ending in .xml) and directories containing .xml files
 -- into a list of .xml files.
 argFiles :: [String] -> IO [String]
@@ -169,6 +171,7 @@ processOneFile src = do
            unlines includeImports ++ 
            "import Text.ParserCombinators.Parsec\n\
            \import Data.List (nub)\n\
+           \import qualified Data.Set as Set\n\
            \import Data.Map (fromList)\n\
            \import Data.Maybe (fromMaybe)\n\n" ++
            render (mkParser syntax) ++ "\n"
@@ -189,7 +192,8 @@ mkParser syntax =
                    text "let style = fromMaybe \"\" $ lookup attr styles" $$
                    text "st <- getState" $$
                    text "let oldCharsParsed = synStCharsParsedInLine st" $$
-                   text "updateState $ \\st -> st { synStCharsParsedInLine = oldCharsParsed + length txt } " $$
+                   text "let prevchar = if null txt then '\\n' else last txt" $$
+                   text "updateState $ \\st -> st { synStCharsParsedInLine = oldCharsParsed + length txt, synStPrevChar = prevchar } " $$
                    text "return (nub [style, attr], txt)")
       parseExpressionInternal = text "parseExpressionInternal = do" $$ (nest 2 $ 
                                   text "context <- currentContext" $$
@@ -217,8 +221,9 @@ mkParser syntax =
                                   , synStLanguage = synLanguage syntax
                                   , synStCurrentLine = ""
                                   , synStCharsParsedInLine = 0
+                                  , synStPrevChar = '\n'
                                   , synStCaseSensitive = synCaseSensitive syntax
-                                  , synStKeywordCaseSensitive = keywordCaseSensitive $ synKeywordAttr syntax
+                                  , synStKeywordCaseSensitive = keywordCaseSensitive (synKeywordAttr syntax)
                                   , synStCaptures = [] }
       initState = text $ "startingState = " ++ show startingState
       sourceLineParser = text "parseSourceLine = manyTill parseExpressionInternal pEndLine"
@@ -244,9 +249,18 @@ mkParser syntax =
                                           (text $ "_ -> return ()")) $$
                                 {- text "pushContext (fromMaybe \"#stay\" $ lookup context lineBeginContexts)" $$ -}
                                 text "lineContents <- lookAhead wholeLine" $$
-                                text "updateState $ \\st -> st { synStCurrentLine = lineContents, synStCharsParsedInLine = 0 }")
+                                text "updateState $ \\st -> st { synStCurrentLine = lineContents, synStCharsParsedInLine = 0, synStPrevChar = '\\n' }")
+      -- we use 'words "blah blah2 blah3"' to keep ghc from inlining the list, which makes compiling take a long time
+      listDef (n, list) = text $ listName n ++ " = Set.fromList $ words $ " ++
+                               show (if keywordCaseSensitive (synKeywordAttr syntax)
+                                        then unwords list
+                                        else map toLower (unwords list))
+      lists = vcat $ map listDef $ synLists syntax
+      regexDef re = text $ compiledRegexName re ++ " = compileRegex " ++ show re
+      regexes = vcat $ map regexDef $ nub $ [parserString x | x <- concatMap contParsers (synContexts syntax),
+                                                              parserType x == "RegExpr", parserDynamic x == False]
   in  vcat $ intersperse (text "") $ [name, exts, mainFunction, parseExpression, mainParser, initState, sourceLineParser, 
-                                      endLineParser, withAttr, styles, parseExpressionInternal, 
+                                      endLineParser, withAttr, styles, parseExpressionInternal, lists, regexes,
                                       defaultAttributes {- , lineBeginContexts -}] ++ contexts ++ [contextCatchAll]
 
 mkAlternatives :: [Doc] -> Doc
@@ -279,9 +293,12 @@ mkSyntaxParser syntax context parser =
             "StringDetect"     -> "pString " ++ show (parserDynamic parser) ++ " " ++ show (parserString parser) 
             "RegExpr"          -> if parserDynamic parser
                                      then "pRegExprDynamic " ++ show (parserString parser)
-                                     else "pRegExpr (compileRegex " ++ show (parserString parser) ++ ")"
-            "keyword"          -> "pKeyword " ++ show (keywordDelims $ synKeywordAttr syntax) ++ " " ++ 
-                                                 show (fromMaybe [] $ lookup (parserString parser) (synLists syntax))
+                                     else "pRegExpr " ++ compiledRegexName (parserString parser)
+            "keyword"          -> "pKeyword " ++ show (keywordDelims $ synKeywordAttr syntax) ++ " " ++ list
+                                     where list = case lookup string (synLists syntax) of
+                                                   Just _   -> listName string
+                                                   Nothing  -> "Set.empty"
+                                           string = parserString parser
             "Int"              -> "pInt"
             "Float"            -> "pFloat"
             "HlCOct"           -> "pHlCOct"
@@ -343,6 +360,18 @@ langNameToModule str =  "Text.Highlighting.Kate.Syntax." ++
     "JavaScript" -> "Javascript"
     x -> x
 
+listName :: String -> String
+listName n = "list_" ++ normalize n
+
+compiledRegexName :: String -> String
+compiledRegexName n = "regex_" ++ normalize n
+
+normalize :: String -> String
+normalize "" = ""
+normalize (x:xs) | isAlphaNum x = x : normalize xs 
+normalize (' ':xs)              = '_':normalize xs
+normalize (x:xs)                = printf "'%2x" (ord x) ++ normalize xs
+
 capitalize :: String -> String
 capitalize (x:xs) = toUpper x : xs
 capitalize [] = []
